import logging
import socket
import urllib.parse
import xml.etree.ElementTree as ET
from collections import OrderedDict

from sans.ethz import Error
from sans.ethz.util import format_mac, httpclient

hostname = 'www.netcenter.ethz.ch'


def uri(relative_uri):
    if not relative_uri.startswith('/'):
        relative_uri = '/{}'.format(relative_uri)
    return 'https://{}{}'.format(hostname, relative_uri)


class NetcenterError(Error):
    """Represents an error that occured while contacting the REST webservice.
    """

    def __init__(self, errors):
        self._errors = errors

    def __str__(self):
        if isinstance(self._errors, (list, tuple)):
            msg = []
            for error in self._errors:
                msg.append(error)
            msg = '\n'.join(msg)
            return msg
        else:
            return str(self._errors)


class NetcenterAuthenticationError(NetcenterError):
    """Represents an authentication error that occured while contacting the REST webservice.
    """

    def __init__(self):
        super(NetcenterAuthenticationError, self).__init__("Authetication failed")


def get_tree_or_raise_error(response, xmlstring):
    """Create and return an element tree or raise an exception if the xml message contains errors.
    """
    if response.status == 401:
        raise NetcenterAuthenticationError()

    xmltree = ET.fromstring(xmlstring)
    if xmltree.tag == 'errors':
        errors = []
        for error in xmltree.findall('error'):
            child_nodes = list(error)
            if child_nodes:
                errors.append(', '.join([child.text for child in child_nodes]))
            else:
                if 'errMsg' in error.attrib:
                    errors.append(error.attrib['errMsg'])
                else:
                    errors.append(error.text)
        raise NetcenterError(errors)
    else:
        return xmltree


def record_to_dict(record):
    d = OrderedDict()
    d.update(record.attrib.copy())
    for child in list(record):
        d[child.tag] = child.text or ''
    return d


def is_phone(entry):
    if 'cable_used_status' in entry:
        return 'Tel' in entry['cable_used_status'].split('/')
    else:
        return 'i_phone_no' in entry


def is_host(entry):
    if 'cable_used_status' in entry:
        return 'DK' in entry['cable_used_status'].split('/')
    else:
        return 'ip' in entry


def is_notused(entry):
    return 'cable_used_status' in entry and entry['cable_used_status'] == 'not used'


class Client(object):
    default_http_headers = {'Content-Type': 'text/xml'}

    def __init__(self, username, password):
        self._http = httpclient.get_http_client(username=username, password=password, host=hostname)

    def GET(self, url, params=None, headers=None):
        """Returns an element tree.
        """
        try:
            if params:
                params = urllib.parse.urlencode(params)
                url = '{}?{}'.format(url, params)
            if headers is None:
                headers = self.default_http_headers.copy()
            logging.debug('uri: %s', url)
            logging.debug('headers: %s', headers)
            response, content = self._http.request(url, headers=headers, auth_without_challenge=True)
            logging.debug('response: %s', response)
            logging.debug('content: %s', content)

            # Create element tree from xml string
            xmltree = get_tree_or_raise_error(response, content)

            return xmltree
        except socket.error as e:
            raise NetcenterError(e)

    def POST(self, url, body=None, params=None, headers=None):
        """Returns an element tree.
        """
        try:
            if params:
                params = urllib.parse.urlencode(params)
                url = '{}?{}'.format(url, params)
            if headers is None:
                headers = self.default_http_headers.copy()
            logging.debug('uri: %s', url)
            logging.debug('headers: %s', headers)
            logging.debug('body: %s', body)
            response, content = self._http.request(url, method='POST', headers=headers, body=body,
                                                   auth_without_challenge=True)
            logging.debug('response: %s', response)
            logging.debug('content: %s', content)

            # Create element tree from xml string
            xmltree = get_tree_or_raise_error(response, content)

            return xmltree
        except socket.error as e:
            raise NetcenterError(e)

    def get_cable_location(self, cable_id):
        """https://netcenter.ethz.ch/netcenter/rest/report/cableLocation01?&cable_id=UTP41972

    <list>
        <record id="0">
            <cable_id>UTP41972</cable_id>
            <start_location>RZ/G/15.2</start_location>
            <end_location>RZ/G/13</end_location>
            <cable_used_status>DK</cable_used_status>
            <Switch>swi-rz-g15-b Gi5/19</Switch>
            <i_phone_no/>
        </record>
    </list>
    """
        url = uri('/netcenter/rest/report/cableLocation01')
        params = {'cable_id': cable_id}
        tree = self.GET(url, params)
        for record in tree.findall('record'):
            yield record_to_dict(record)

    def get_subnet_info(self, vlan_subnet_ip):
        """https://netcenter.ethz.ch/netcenter/rest/report/vlanSubnetIp01?&input=129.132.12.0

    <list>
        <record id="0">
            <vlan>Cx::781</vlan>
            <vlan_name>inf-icos.cx.dock.1</vlan_name>
            <ip_subnet>129.132.12.0/24</ip_subnet>
            <subnet_name>inf-icos-cx-dock-1</subnet_name>
        </record>
    </list>
        """
        if not isinstance(vlan_subnet_ip, (list, tuple)):
            vlan_subnet_ip = [vlan_subnet_ip]

        url = uri('/netcenter/rest/report/vlanSubnetIp01')

        for query in vlan_subnet_ip:
            params = {'input': query}
            tree = self.GET(url, params)

            for record in tree.findall('record'):
                subnet = record_to_dict(record)
                yield subnet

    def get_host_ip(self, _hostname):
        """POST https://www.netcenter.ethz.ch/netcenter/nameToIP/loadNameToIPList
    <query>
      <nameToIP>ipctl.icos.ethz.ch</nameToIP>
      <ipOrAlias/>
      <example/>
    </query>

    <<
    <?xml version="1.0"?>
    <list>
      <nameToIP id="2588265" status="unchanged">
        <errors>
        </errors>
        <locked>N</locked>
        <isv6>false</isv6>
        <ip>129.132.12.49</ip>
        <ipSubnet>129.132.12.0</ipSubnet>
        <prefix></prefix>
        <name>ipctl</name>
        <subdomainName>icos.ethz.ch</subdomainName>
        <subdomain>2481782</subdomain>
        <forward>Y</forward>
        <reverse>Y</reverse>
        <ttl>1200</ttl>
        <dhcp>N</dhcp>
        <dhcpMac></dhcpMac>
        <ddns>N</ddns>
        <isgId>4136303</isgId>
        <isgGroup>inf-icos</isgGroup>
        <subnetIsgId>4136303</subnetIsgId>
        <lastDetection>2008-12-17 12:06</lastDetection>
        <alias>Y</alias>
        <lastChange>2010-05-11 08:28</lastChange>
        <lastChangeBy>idwssa</lastChangeBy>
      </nameToIP>
    </list>

    <query>
      <nameToIP>ipctl.icos.ethz.ch</nameToIP>
      <ipOrAlias/>
      <example/>
    </query>

        """
        # FIXME: need some new webservice thinggy to get this
        return None
        url = uri('/netcenter/nameToIP/loadNameToIPList')
        body = '<query><nameToIP>{}</nameToIP></query>'.format(_hostname)
        tree = self.POST(url, body=None, params=None, headers=None)

        try:
            ip = tree.find('nameToIP/ip').text
        except AttributeError:
            ip = None
        return None

    def get_host_info(self, ip_name_mac_cable):
        """https://netcenter.ethz.ch/netcenter/rest/report/hostInfo01?&ip_name_mac_cable=129.132.12.6

    <list>
        <record id="0">
            <mac>001b.fca2.c39f</mac>
            <ip>129.132.12.6</ip>
            <ip_name>coeus.inf.ethz.ch</ip_name>
            <ip_subnet>129.132.12.0</ip_subnet>
            <vpz>inf-icos</vpz>
            <vlan>Cx::781</vlan>
            <vlan_name>inf-icos.cx.dock.1</vlan_name>
            <vtp>Cx</vtp>
            <switch_ip>172.20.202.7</switch_ip>
            <switch_name>swi-cab-e55-2-a</switch_name>
            <switch_port>Gi5/5</switch_port>
            <macro_name>fix V3</macro_name>
            <cable_id>UTP128536</cable_id>
            <end_location>CAB/E/48</end_location>
            <created>11-12-2007 15:20:15.000000</created>
            <last_seen>04-02-2009 13:01:40.000000</last_seen>
        </record>
    </list>
        """
        if not isinstance(ip_name_mac_cable, (list, tuple)):
            ip_name_mac_cable = [ip_name_mac_cable]

        url = uri('/netcenter/rest/report/hostInfo01')

        for query in ip_name_mac_cable:
            params = {'ip_name_mac_cable': query}
            tree = self.GET(url, params)

            for record in tree.findall('record'):
                host = record_to_dict(record)
                if 'mac' in host:
                    host['mac'] = format_mac(host['mac'])
                yield host

    def get_profile_info(self, net_sup):
        url = uri('/netcenter/rest/report/profileInfo')
        params = dict.fromkeys(
            'searchVpz searchRealm searchV6Subnet searchVlan searchProfile searchV4Subnet'.split(' '), '')
        params['searchNetsup'] = net_sup
        tree = self.GET(url, params)
        for record in tree.findall('record'):
            profile = record_to_dict(record)
            yield profile

    def search_by_ip_name_mac_cable(self, ip_name_mac_cable):
        if not isinstance(ip_name_mac_cable, (list, tuple)):
            ip_name_mac_cable = [ip_name_mac_cable]

        for query in ip_name_mac_cable:
            found_host = False
            for host in self.get_host_info(query):
                # host
                found_host = True
                yield host
            if not found_host:
                # phone or non existent
                if query.startswith('UTP') or query.startswith('PAT'):
                    for device in self.get_cable_location(query):
                        yield device

    def search_by_vlan(self, vlans, rooms=None):
        """https://netcenter.ethz.ch/netcenter/rest/report/utpsInVlan01?&vlan=Cx::781

    <list>
        <record id="0">
            <vlan>Cx::781</vlan>
            <switch_name>swi-cab-f70-4-a</switch_name>
            <switch_port>Gi4/46</switch_port>
            <macro_name>fix V3</macro_name>
            <cable_id>UTP130217</cable_id>
            <end_location>CAB/0/DUMMY</end_location>
            <patch_date>08-04-2008 14:04:55.000000</patch_date>
            <last_up/>
        </record>
        <record id="1">
            <vlan>Cx::781</vlan>
            <switch_name>swi-cab-d56-a</switch_name>
            <switch_port>Gi7/17</switch_port>
            <macro_name>fix V3</macro_name>
            <cable_id>PAT132265</cable_id>
            <end_location>CAB/D/56</end_location>
            <patch_date>11-06-2008 11:42:50.000000</patch_date>
            <last_up>06-02-2009 14:46:09.000000</last_up>
        </record>
    </list>
        """
        if not isinstance(vlans, (list, tuple)):
            vlans = [vlans]

        if rooms and not isinstance(rooms, (list, tuple)):
            rooms = [rooms]

        url = uri('/netcenter/rest/report/utpsInVlan01')

        for query in vlans:
            if '::' not in query:
                query = '%::{}'.format(query)
            params = {'vlan': query}
            tree = self.GET(url, params)

            for record in tree.findall('record'):
                vlan = record.find('vlan').text
                cable_id = None
                if rooms:
                    room = record.find('end_location')
                    if room.text:
                        for r in rooms:
                            if room.text == r:
                                cable_id = record.find('cable_id').text
                else:
                    cable_id = record.find('cable_id').text
                if cable_id:
                    for device in self.search_by_ip_name_mac_cable(cable_id):
                        if 'vlan' not in device:
                            device['vlan'] = vlan
                        yield device

    def search_by_room(self, rooms, vlans=None):
        """https://netcenter.ethz.ch/netcenter/rest/report/eloc2utp01?&eLoc=CAB/E/48

    <list>
        <record id='8'>
            <start_location>CAB/E/55.2</start_location>
            <cable_id>UTP128531</cable_id>
            <end_location>CAB/E/48</end_location>
            <cable_used_status>Tel</cable_used_status>
            <i_phone_no>22524</i_phone_no>
            <switch_ip/>
            <switch_name/>
            <switch_port/>
            <macro_name/>
            <vlan/>
            <patch_date></patch_date>
            <last_up></last_up>
            <always_patched>Nein</always_patched>
        </record>
        <record id="dummy">
            <start_location>CAB/E/55.2</start_location>
            <cable_id>UTP128532</cable_id>
            <end_location>CAB/E/48</end_location>
            <cable_used_status>not used</cable_used_status>
            <i_phone_no/>
            <switch_ip/>
            <switch_name/>
            <switch_port/>
            <macro_name/>
            <vlan/>
            <patch_date/>
            <last_up/>
            <always_patched>Nein</always_patched>
        </record>
        <record id='13'>
            <start_location>CAB/E/55.2</start_location>
            <cable_id>UTP128536</cable_id>
            <end_location>CAB/E/48</end_location>
            <cable_used_status>DK</cable_used_status>
            <i_phone_no/>
            <switch_ip>172.20.202.7</switch_ip>
            <switch_name>swi-cab-e55-2-a</switch_name>
            <switch_port>Gi5/5</switch_port>
            <macro_name>fix V3</macro_name>
            <vlan>Cx::781</vlan>
            <patch_date>03-05-2007 12:08:20.000000</patch_date>
            <last_up>04-02-2009 16:35:52.000000</last_up>
            <always_patched>Nein</always_patched>
        </record>
    </list>
        """
        if not isinstance(rooms, (list, tuple)):
            rooms = [rooms]

        if vlans and not isinstance(vlans, (list, tuple)):
            vlans = [vlans]

        url = uri('/netcenter/rest/report/eloc2utp01')

        for query in rooms:
            params = {'eLoc': query, 'fibIncl': 'true', 'balanceLineIncl': 'true', 'planningIncl': 'false'}
            tree = self.GET(url, params)

            for record in tree.findall('record'):
                entry = record_to_dict(record)
                if vlans:
                    entry_not_in_vlan = True
                    if 'vlan' in entry and entry['vlan']:
                        for vlan in vlans:
                            if entry['vlan'].endswith(vlan):
                                entry_not_in_vlan = False
                                break
                    if entry_not_in_vlan:
                        continue

                if is_host(entry):
                    if is_phone(entry):
                        # voip phone
                        yield entry
                    for host in self.get_host_info(entry['cable_id']):
                        yield host
                else:
                    yield entry
