import logging

import sans.cmdln
from sans.cmdln import *


def echo(msg, *args):
    if args:
        if isinstance(args[0], dict):
            msg = msg.format(**args[0])
            _log(msg)
            return
        else:
            a = []
            for arg in args:
                a.append(arg)
            args = tuple(a)
        msg = msg.format(*args)
    _log(str(msg))


def _log(s):
    from os.path import expanduser
    fout = open(expanduser('/tmp/bashcpln.log'), 'a')
    fout.write(str(s) + '\n')
    fout.close()


# Recipe: module_from_path (1.0.1+)
def _module_from_path(path):
    import imp, os, sys
    path = os.path.expanduser(path)
    dir = os.path.dirname(path) or os.curdir
    name = os.path.splitext(os.path.basename(path))[0]
    sys.path.insert(0, dir)
    try:
        iinfo = imp.find_module(name, [dir])
        return imp.load_module(name, *iinfo)
    finally:
        sys.path.remove(dir)


def _get_bash_cplns2(shell, cmd_name, token, preceding_token):
    echo('get_cplns({}, {}, {}, {})', shell, cmd_name, token, preceding_token)
    comp_line = os.environ["COMP_LINE"]
    comp_point = int(os.environ["COMP_POINT"])
    echo("COMP_LINE: {}", comp_line)
    echo("COMP_POINT: {}", comp_point)

    cmd_map = shell._get_canonical_map()
    # key = alias, value = command
    echo('cmd_map: {}', cmd_map)

    names = shell.get_names()
    cmdnames = {}
    for attr in shell.get_names():
        if attr.startswith("do_"):
            cmdnames[attr[3:]] = True
    echo('cmdnames: {}', cmdnames)

    del cmd_map["EOF"]

    # Determine if completing the sub-command name.
    parts = comp_line[:comp_point].split(None, 1)
    echo(parts)
    if len(parts) == 1 or not (' ' in parts[1] or '\t' in parts[1]):
        # TODO: if parts[1].startswith('-'): handle top-level opts
        echo("complete sub-command names")
        matches = {}
        for name, canon_name in cmd_map.items():
            if name.startswith(token):
                matches[name] = canon_name
        if not matches:
            return []
        elif len(matches) == 1:
            return matches.keys()
        elif len(set(matches.values())) == 1:
            return [matches.values()[0]]
        else:
            return matches.keys()

    # Otherwise, complete options for the given sub-command.
    # TODO: refine this so it does the right thing with option args
    if token.startswith('-'):
        cmd_name = comp_line.split(None, 2)[1]
        try:
            cmd_canon_name = cmd_map[cmd_name]
        except KeyError:
            return []
        handler = shell._get_cmd_handler(cmd_canon_name)
        optparser = getattr(handler, "optparser", None)
        if optparser is None:
            optparser = sans.cmdln.SubCmdOptionParser()
        opt_strs = []
        for option in optparser.option_list:
            for opt_str in option._short_opts + option._long_opts:
                if opt_str.startswith(token):
                    opt_strs.append(opt_str)
        return opt_strs

    return []


class RemainderPreservingOptionParser(sans.cmdln.SubCmdOptionParser):
    def parse_args(self, args=None, values=None):
        """
        parse_args(args : [string] = sys.argv[1:],
                   values : Values = None)
        -> (values : Values, args : [string], rargs : [string])

        See OptionParser.parse_args help for details.

        The only change to that is the we allow the return of the so called rargs
        ("remaining" or "right-hand") untouched.

        Beware that this changes the signature of the return value which
        now returns 3 instead of 2 values. 
        """
        rargs = self._get_args(args)
        if values is None:
            values = self.get_default_values()

        # Store the halves of the argument list as attributes for the
        # convenience of callbacks:
        #   rargs
        #     the rest of the command-line (the "r" stands for
        #     "remaining" or "right-hand")
        #   largs
        #     the leftover arguments -- ie. what's left after removing
        #     options and their arguments (the "l" stands for "leftover"
        #     or "left-hand")
        self.rargs = rargs
        self.largs = largs = []
        self.values = values

        try:
            stop = self._process_args(largs, rargs, values)
        except (optparse.BadOptionError, optparse.OptionValueError) as err:
            self.error(str(err))

        values, args = self.check_values(values, largs)
        return values, args, rargs


class Command(sans.cmdln.Cmdln, object):
    option_debug = True
    option_quiet = True
    option_interactive = False
    use_history = True

    def __init__(self, *args, **kwargs):
        sans.cmdln.Cmdln.__init__(self, *args, **kwargs)
        self.log = logging.getLogger(self.name)

    def _do_loglevel(self, subcmd, opts, *args):
        """hidden command to set or show the logging level"""
        level = logging.root.getEffectiveLevel()
        if args:
            level_name = args[0]
            level = getattr(logging, level_name.upper())
            logging.root.setLevel(level)
            self.log.setLevel(level)
        print(logging.getLevelName(level))

    def _do_debug(self, subcmd, opts, *args):
        """hidden command to enable/disable debug logging"""
        action = 'on'
        if args:
            action = args[0]
        if action in ('on', '1'):
            level_name = 'debug'
        elif action in ('off', '0'):
            level_name = 'error'
        self._do_loglevel(subcmd, opts, level_name)

    def _set_debug(self, option, opt_str, value, parser, *args, **kwargs):
        logging.root.setLevel(logging.DEBUG)
        self.log.setLevel(logging.DEBUG)
        parser.values.debug = True

    def _set_quiet(self, option, opt_str, value, parser, *args, **kwargs):
        logging.root.setLevel(logging.ERROR)
        self.log.setLevel(logging.ERROR)
        parser.values.quiet = True

    def get_optparser(self):
        parser = sans.cmdln.Cmdln.get_optparser(self)
        parser.set_defaults(debug=False, quiet=False)
        if self.option_debug:
            parser.add_option('-d', '--debug', action='callback', callback=self._set_debug)
        if self.option_quiet:
            parser.add_option('-q', '--quiet', action='callback', callback=self._set_quiet,
                              help='disable all optional output')
        if self.option_interactive:
            parser.add_option('-i', '--interactive', action='store_true',
                              help='run interactively as a shell')

        #        parser.add_option('-d', '--debug', action='store_true',
        #                      help='print info useful for debugging')
        #        parser.add_option('-n', '--dry-run', action='store_true',
        #                      help='just print out what would be done')
        return parser

    def preloop(self):
        if self.use_history and self.use_rawinput and self.completekey:
            try:
                # Configure readline history
                import readline
                if hasattr(readline, 'read_history_file'):
                    # Don't clutter the users home directory
                    config_home = os.environ.get('XDG_CONFIG_HOME', '~/.config')
                    config_dir = os.path.expanduser(os.path.join(config_home, 'sans'))
                    if not os.path.isdir(config_dir):
                        os.mkdir(config_dir, 0o700)
                    history_file = os.path.join(config_dir, '{}.history'.format(self.name))
                    logging.debug('Using readline history file: %s', history_file)

                    def save_history(history_file):
                        readline.write_history_file(history_file)

                    import atexit
                    atexit.register(save_history, history_file)

                    if os.path.isfile(history_file):
                        readline.read_history_file(history_file)

            except ImportError:
                pass
            except IOError as e:
                logging.warning(e)

    def main(self, argv=None, loop=sans.cmdln.LOOP_NEVER):
        """A possible mainline handler for a script, like so:

            import cmdln
            class MyCmd(cmdln.Cmdln):
                name = "mycmd"
                ...

            if __name__ == "__main__":
                MyCmd().main()

        By default this will use sys.argv to issue a single command to
        'MyCmd', then exit. The 'loop' argument can be use to control
        interactive shell behaviour.

        Arguments:
            "argv" (optional, default sys.argv) is the command to run.
                It must be a sequence, where the first element is the
                command name and subsequent elements the args for that
                command.
            "loop" (optional, default LOOP_NEVER) is a constant
                indicating if a command loop should be started (i.e. an
                interactive shell). Valid values (constants on this module):
                    LOOP_ALWAYS     start loop and run "argv", if any
                    LOOP_NEVER      run "argv" (or .emptyline()) and exit
                    LOOP_IF_EMPTY   run "argv", if given, and exit;
                                    otherwise, start loop
        """
        # 20090601/asteven: if outputting to a pipe, imply --quiet
        # FIXME: broken by writing to parser.values in _set_quiet
        # if not sys.stdout.isatty():
        #    self._set_quiet(None, None, None, self.optparser)

        if argv is None:
            argv = sys.argv
        else:
            argv = argv[:]  # don't modify caller's list

        self.optparser = self.get_optparser()
        if self.optparser:  # i.e. optparser=None means don't process for opts
            try:
                self.options, args = self.optparser.parse_args(argv[1:])
            except sans.cmdln.CmdlnUserError as ex:
                msg = "{}: {}\nTry '{} help' for info.\n".format(self.name, ex, self.name)
                self.stderr.write(self._str(msg))
                self.stderr.flush()
                return 1
            except sans.cmdln.StopOptionProcessing as ex:
                return 0
        else:
            self.options, args = None, argv[1:]
        self.postoptparse()

        # 20090601/asteven: allow interactive mode to be set by providing a command line option
        if self.option_interactive and self.options.interactive:
            loop = sans.cmdln.LOOP_ALWAYS

        if loop == sans.cmdln.LOOP_ALWAYS:
            if args:
                self.cmdqueue.append(args)
            return self.cmdloop()
        elif loop == sans.cmdln.LOOP_NEVER:
            if args:
                return self.cmd(args)
            else:
                return self.emptyline()
        elif loop == sans.cmdln.LOOP_IF_EMPTY:
            if args:
                return self.cmd(args)
            else:
                return self.cmdloop()

    def cmdexc(self, argv):
        """Called if an exception is raised in any of precmd(), onecmd(),
        or postcmd(). If True is returned, the exception is deemed to have
        been dealt with. Otherwise, the exception is re-raised.

        The default implementation handles CmdlnUserError's, which
        typically correspond to user error in calling commands (as
        opposed to programmer error in the design of the script using
        cmdln.py).
        """
        import sys
        type, exc, traceback = sys.exc_info()
        if isinstance(exc, (sans.cmdln.CmdlnUserError, AssertionError)):
            msg = "{} {}: {}\nTry '{} help {}' for info.\n".format(self.name, argv[0], exc, self.name, argv[0])
            self.stderr.write(self._str(msg))
            self.stderr.flush()
            return True

    def parse_cmd_args(self, handler, argv):
        """returns (dict,list)
        raises StopOptionProcessing
        """
        try:
            optparser = handler.optparser
        except AttributeError:
            optparser = handler.__func__.optparser = sans.cmdln.SubCmdOptionParser()
        assert isinstance(optparser, sans.cmdln.SubCmdOptionParser)
        optparser.set_cmdln_info(self, argv[0])
        return optparser.parse_args(argv[1:])

    def _dispatch_cmd(self, handler, argv):
        """Introspect sub-command handler signature to determine how to
        dispatch the command. The raw handler provided by the base
        'RawCmdln' class is still supported:

            def do_foo(self, argv):
                # 'argv' is the vector of command line args, argv[0] is
                # the command name itself (i.e. "foo" or an alias)
                pass

        In addition, if the handler has more than 2 arguments option
        processing is automatically done (using optparse):

            @cmdln.option('-v', '--verbose', action='store_true')
            def do_bar(self, subcmd, opts, *args):
                # subcmd = <"bar" or an alias>
                # opts = <an optparse.Values instance>
                if opts.verbose:
                    print "lots of debugging output..."
                # args = <tuple of arguments>
                for arg in args:
                    bar(arg)

        TODO: explain that "*args" can be other signatures as well.

        The `cmdln.option` decorator corresponds to an `add_option()`
        method call on an `optparse.OptionParser` instance.

        You can declare a specific number of arguments:

            @cmdln.option('-v', '--verbose', action='store_true')
            def do_bar2(self, subcmd, opts, bar_one, bar_two):
                #...

        and an appropriate error message will be raised/printed if the
        command is called with a different number of args.
        """
        co_argcount = handler.__func__.__code__.co_argcount
        if co_argcount == 2:  # handler ::= do_foo(self, argv)
            return handler(argv)
        elif co_argcount >= 3:  # handler ::= do_foo(self, subcmd, opts, ...)
            try:
                arguments = self.parse_cmd_args(handler, argv)
            except sans.cmdln.StopOptionProcessing:
                # TODO: this doesn't really fly for a replacement of
                #      optparse.py behaviour, does it?
                return 0  # Normal command termination

            try:
                if len(arguments) == 3:
                    opts, args, rargs = arguments
                    self.log.debug('%s %s %s %s', argv[0], opts, args, rargs)
                    return handler(argv[0], opts, args, rargs)
                else:
                    opts, args = arguments
                    self.log.debug('%s %s %s', argv[0], opts, args)
                    return handler(argv[0], opts, *args)
            except TypeError as ex:
                self.log.error(ex)
                return handler(argv[0], opts, *args)
            except TypeError as ex:
                # Some TypeError's are user errors:
                #   do_foo() takes at least 4 arguments (3 given)
                #   do_foo() takes at most 5 arguments (6 given)
                #   do_foo() takes exactly 5 arguments (6 given)
                # Raise CmdlnUserError for these with a suitably
                # massaged error message.
                import sys
                tb = sys.exc_info()[2]  # the traceback object
                if tb.tb_next is not None:
                    # If the traceback is more than one level deep, then the
                    # TypeError do *not* happen on the "handler(...)" call
                    # above. In that we don't want to handle it specially
                    # here: it would falsely mask deeper code errors.
                    raise
                msg = ex.args[0]
                match = sans.cmdln._INCORRECT_NUM_ARGS_RE.search(msg)
                if match:
                    msg = list(match.groups())
                    msg[1] = int(msg[1]) - 3
                    if msg[1] == 1:
                        msg[2] = msg[2].replace("arguments", "argument")
                    msg[3] = int(msg[3]) - 3
                    msg = ''.join(map(str, msg))
                    raise sans.cmdln.CmdlnUserError(msg)
                else:
                    raise
        else:
            raise sans.cmdln.CmdlnError("incorrect argcount for {}(): takes {}, must "
                                        "take 2 for 'argv' signature or 3+ for 'opts' "
                                        "signature".format(handler.__name__, co_argcount))

    def _do_bashcomp(self, argv):
        """Usage: complete -C "netinfo bashcomp" netinfo
        """
        if False:
            echo('sys.argv: {}', sys.argv)
            if len(sys.argv) == 6:
                for cpln in _get_bash_cplns(*sys.argv[1:]):
                    print(cpln)

        else:
            echo('--')
            echo('argv: {}', argv)
            if len(argv) == 4:
                args = argv[:]
                args[0] = self
                for cpln in _get_bash_cplns2(*args):
                    print(cpln)


# --- Deorators -----

def option_parser(option_parser_class):
    """Decorator for using a different OptionParser for a method.

    Note that this must be the last decorator used on the class, so it is called first.

    @some_decorator
    @option_parser(MyOptionParserClass)
    def foo(self):
        pass
    """

    def decorate(f):
        f.optparser = option_parser_class()
        return f

    return decorate


"""Don't parse remaining arguments after -- 
instead return them untouched.
"""
preserve_remaining_arguments = option_parser(RemainderPreservingOptionParser)
