#!/usr/bin/env python3
# $Id:$
#
# Query the ETHZ LDAP server and/or a local ~/.phonebook file for addressbook entries.
#
# (c) 2006, Steven Armstrong, ETH Zurich
#
#

import os
import re
import sys

try:
    import ldap
except ImportError:
    print('You need to install the python ldap bindings to use this command.')
    print('On Gentoo run: /usr/bin/emerge python3-ldap -va')
    print('On Debian run: apt-get install python3-ldap')
    sys.exit(1)

__version__ = '0.1.3'

# Column Identifiers:
#  n: Name
#  p: Phone
#  a: Address
#  r: Room
#  e: Email
#  d: Department

default_columns = 'nrdpe'
local_phonebook_file = '.phonebook'
ldap_url = "ldap://ldap.ethz.ch/"
base_staff = "ou=staff,o=ethz,c=ch"
base_student = "ou=student,o=ethz,c=ch"

phone_re = re.compile('^([\d\s*]+)$')
room_re = re.compile('^(([^\d\s\/]+)[\s\/]?([^\d\s]{1})[\s\/]?(\d+(?:.\d+)*))$')
email_re = re.compile(
    r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"  # dot-atom
    r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*"'  # quoted-string
    r')@(?:[A-Z0-9-]+\.)+[A-Z]{2,6}$', re.IGNORECASE)  # domain

VERBOSITY = 0


def info(msg, *args):
    if VERBOSITY > 0:
        print(msg.format(args))


def debug(msg, *args):
    if VERBOSITY > 1:
        print(msg.format(args))


def error(msg, *args):
    print(msg.format(args), file=sys.stderr)


def echo(msg, *args):
    if args:
        if isinstance(args[0], dict):
            print(msg.format(**args[0]))
            return
        else:
            a = []
            for arg in args:
                a.append(arg)
            args = tuple(a)
    print(msg.format(args))


class ColumnModel(object):
    fields = ['name', 'phone', 'address', 'room', 'email', 'department']
    labels = ['Name', 'Phone', 'Address', 'Room', 'Email', 'Dept']
    all_columns = 'npared'

    def __init__(self, columns=None):
        if columns:
            self.columns = dict(zip(list(columns), self.fields))
        else:
            self.columns = dict(zip(list(self.all_columns), self.fields))
        self.labels = dict(zip(self.fields, self.labels))
        # initially set columns widths to width of labels
        self.widths = dict(zip(self.fields, [len(label) for label in self.labels]))

    def update_widths(self, entry):
        for k, v in entry.items():
            if v:
                l = len(v)
                if k in self.widths and l > self.widths[k]:
                    self.widths[k] = l

    def get_widths(self):
        return dict([(k, v + 2) for k, v in self.widths.items()])

    def get_total_width(self, columns):
        width = 0
        widths = self.get_widths()
        for c in list(columns):
            width += widths[self.columns[c]]
        return width

    def make_format(self, columns):
        _format = []
        widths = self.get_widths()
        for c in columns:
            name = self.columns[c]
            _format.append('{' + name + ': <' + str(widths[name]) + '}')
        _format = ''.join(_format)
        return _format

    def make_entry(self, columns, line):
        fields = [field.strip().rstrip('\r\n') for field in line.split('\t')]
        entry = dict.fromkeys(self.fields, '')
        for k, v in zip(columns, fields):
            name = self.columns[k]
            entry[name] = v
        return entry


def search_ldap(_options, _args, columns, url, base):
    entries = []

    scope = ldap.SCOPE_SUBTREE
    # ret = None # get whole record
    ret = ['cn', 'roomNumber', 'ethzDepartment', 'telephoneNumber', 'mail', 'postalAddress']
    timeout = 0

    _filter = ['(&']
    if _options.room:
        _filter.append('(&(roomNumber=*{}*))'.format(_options.room))
    elif _options.phone:
        _filter.append('(&(telephoneNumber=*{}*))'.format(_options.phone))
    elif _options.email:
        _filter.append('(&(mail=*{}*))'.format(_options.email))
    else:
        for arg in _args:
            if _options.approx:
                q = '(|(sn~=%(arg)s)(givenName~=%(arg)s)(mail~=%(arg)s))'
            else:
                q = '(|(sn=*%(arg)s*)(givenName=*%(arg)s*)(mail=*%(arg)s*))'
            _filter.append(q % {'arg': arg})

    _filter.append(')')
    _filter = ''.join(_filter)

    debug('name: {}, phone: {}, room: {}, email: {}', _options.name, _options.phone, _options.room, _options.email)
    debug('filter: {}', _filter)

    try:
        info('Connecting to: {}', url)
        l = ldap.initialize(url)
        l.simple_bind_s('', '')

        info('Searching...')

        try:
            result_id = l.search(base, scope, _filter, ret)
            debug('result_id: {}', result_id)

            while True:
                result_type, result_data = l.result(result_id, timeout)
                debug('result_type: {}', result_type)
                debug('result_data: {}', result_data)

                if result_data:
                    if result_type == ldap.RES_SEARCH_ENTRY:
                        entry = {
                            'name': result_data[0][1]['cn'][0].decode(),
                            'room': result_data[0][1].get('roomNumber', [b''])[0].decode(),
                            'department': result_data[0][1].get('ethzDepartment', [''])[0],
                            'phone': result_data[0][1].get('telephoneNumber', [b''])[0].decode(),
                            'email': result_data[0][1].get('mail', [b''])[0].decode(),
                            'address': result_data[0][1].get('postalAddress', [b''])[0].decode()
                        }
                        # change room delimiter from space to / for consistency with other tools
                        entry['room'] = entry['room'].replace(' ', '/')
                        columns.update_widths(entry)
                        entries.append(entry)
                else:
                    break

            return entries

        except ldap.LDAPError as error_message:
            error('Search failed: {}', error_message)

    except ldap.LDAPError as error_message:
        error('Connection failed: {}', error_message)


def search_entry(_options, entry):
    debug('search_entry: {}', _options.name)
    key = None
    # Narrow down search to room, phone or email
    if _options.room:
        key = 'room'
    elif _options.phone:
        key = 'phone'
    elif _options.email:
        key = 'email'
    elif _options.name:
        key = 'name'
    if key:
        value = getattr(_options, key)
        debug('searching by: {}={}', key, value)
        return value.lower() in entry[key].lower()

    # Default to true as we already checked if the line itself contains the query
    return True


def search_file(_options, _args, columns, filename='.phonebook'):
    query = ' '.join(_args).lower()

    f = open(filename, 'r')
    entries = []
    cols = ''
    for line in f:
        # Search for column definition marker
        if line.startswith('#--columns:'):
            cols = line[11:].strip()
        elif not line.startswith('#') and line is not '':
            if not cols:
                global default_columns
                cols = default_columns
            if query in line.lower():
                # Only make entries if the line contains the query
                entry = columns.make_entry(cols, line)
                if search_entry(_options, entry):
                    columns.update_widths(entry)
                    entries.append(entry)
    f.close()
    return entries, cols


def search_local_phonebook(_options, _args, local_phonebook):
    columns = ColumnModel()
    results_local, cols_local = search_file(_options, _args, columns, local_phonebook)
    if results_local:
        _format = columns.make_format(cols_local)
        width = columns.get_total_width(cols_local)
        # Make sure width is dividable through 2
        width -= width % 2
        print_header('local', width, _format, columns.labels)
        print_data(results_local, _format)


def get_local_phonebook():
    phone_book_filename = '.phonebook'
    user_phone_book_filename = os.path.expanduser('~/{}'.format(phone_book_filename))
    phone_book_file = None
    if os.path.isfile(phone_book_filename):
        phone_book_file = os.path.abspath(phone_book_filename)
    elif os.path.isfile(user_phone_book_filename):
        phone_book_file = user_phone_book_filename
    return phone_book_file


def create_phonebook_example():
    phonebook_file = os.path.expanduser('~/.phonebook')
    if not os.path.exists(phonebook_file):
        target = open(phonebook_file, 'w')
        print('Creating local phonebook file at {}.'.format(phonebook_file))
    else:
        target = sys.stdout
        print('Local phonebook file {} already exists. Writing example to stdout instead.'.format(phonebook_file))
    target.write("""# Possible columns
# n: Name
# p: Phone
# a: Address
# e: Email
# r: Room
# d: Department
#
# Columns are seperated by tabs.
#
# Specify the columns used in this file:
#--columns: npae
Frank Sinatra\t099 999 99 99\tCloud 99, 9999 Heaven\tfrank@heaven.com
""")
    target.flush()
    if target != sys.stdout:
        target.close()


def print_header(title, width, _format, labels):
    num_dashes = (width - len(title) - 7)
    header = '----- {} {}'.format(title, '-' * num_dashes)
    echo(header)
    echo(_format, labels)
    echo('-' * len(header))


def print_data(result_set, _format):
    for r in result_set:
        echo(_format, r)
    echo('')


def main():
    _options, _args = parse_args()
    if _options.create_phonebook_example:
        create_phonebook_example()
        return

    query = ' '.join(_args)

    if None in (_options.name, _options.room, _options.phone, _options.email):
        if phone_re.search(query):
            # Searching for phone number
            # Phone numbers are stored without spaces in LDAP
            _options.phone = ''.join(_args)
        elif room_re.search(query):
            # Searching for room number
            # Room names are stored with spaces in LDAP
            m = room_re.match(query)
            query = ' '.join(m.groups()[1:])
            _options.room = query
        # elif email_re.search(query):
        elif '@' in query:
            # Searching for email
            _options.email = query
        elif not _options.name:
            # Searching for name/text
            _options.name = query

    debug('options: {}', _options)
    debug('args: {}', _args)

    # Search local .phonebook file
    local_phonebook = get_local_phonebook()
    if local_phonebook:
        search_local_phonebook(_options, _args, local_phonebook)
    if _options.only_search_local:
        return

    columns = ColumnModel()

    # Search ETHZ LDAP for students
    if _options.search_students:
        results_student = search_ldap(_options, _args, columns, ldap_url, base_student)

    # Search ETHZ LDAP for staff
    results_staff = search_ldap(_options, _args, columns, ldap_url, base_staff)

    _format = columns.make_format(_options.columns)
    debug('format: {}', _format)

    width = columns.get_total_width(_options.columns)
    # Make sure width is dividable through 2
    width -= width % 2
    debug('width: {}', width)
    labels = columns.labels

    if results_staff:
        print_header('staff', width, _format, labels)
        print_data(results_staff, _format)

    if _options.search_students:
        if results_student:
            print_header('students', width, _format, labels)
            print_data(results_student, _format)


def parse_args():
    import optparse
    usage = 'Usage: %prog [options] text, phone, room or email'
    usage = """{}

Examples:
  ethtel Steven Armstrong
  ethtel 044 632 25 24
  ethtel 22524
  ethtel CAB/E/48
  ethtel CAB E 48
  ethtel steven.armstrong@inf.ethz.ch
  ethtel -l My local phonebook entry""".format(usage)
    parser = optparse.OptionParser(usage=usage, version="%%prog {}".format(__version__))
    parser.set_defaults(verbose=0)
    parser.add_option('-n', '--name',
                      type='string', dest='name',
                      default=None,
                      help='search by name, e.g. John Wayne')
    parser.add_option('-r', '--room',
                      type='string', dest='room',
                      default=None,
                      help='search by room, e.g. ETZ/F/83')
    parser.add_option('-p', '--phone',
                      type='string', dest='phone',
                      default=None,
                      help='search by phone, e.g. 5 88 99')
    parser.add_option('-e', '--email',
                      type='string', dest='email',
                      default=None,
                      help='search by email')
    parser.add_option('-a', '--approx',
                      action='store_true', dest='approx',
                      help='approximate match (slow)')
    parser.add_option('-s', '--students',
                      action='store_true', dest='search_students',
                      help='also search for students (slow)')
    parser.add_option('-l', '--local',
                      action='store_true', dest='only_search_local',
                      help='only search local .phonebook (fast)')
    global default_columns
    parser.add_option('-c', '--columns',
                      type='string', dest='columns',
                      default=default_columns,
                      help='columns to show in output')
    parser.add_option('-v', '--verbose',
                      action='count', dest='verbose',
                      help='be verbose, add multiple times for more verboseness, e.g. -vv')
    parser.add_option('-q', '--quiet',
                      action='store_true', dest='quiet',
                      help='be quiet, overrides -v')

    parser.add_option('-x', '--phonebook-example',
                      action='store_true', dest='create_phonebook_example',
                      help='create a local phonebook example')

    _options, _args = parser.parse_args()
    if _options.quiet:
        _options.verbose = 0

    if not _options.create_phonebook_example:
        if None in (_options.name, _options.room, _options.phone, _options.email) and not _args:
            parser.error('No search query specified. -h for help.')

    global VERBOSITY
    VERBOSITY = _options.verbose

    return _options, _args


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(0)
