#!/usr/bin/env python3
#
# Host control. Perform actions on one or more hosts.
#
# (c) 2009, Steven Armstrong, ETH Zurich
#


"""
$ cat ~/.hostctl.conf
[global]
xterm = urxvt

[csb]
hosts = "csb{2,6,8-13}.inf.ethz.ch"

[cbrg]
hosts = "csb{2,7}.inf.ethz.ch"
groups = linneus-all

[linneus-lower]
hosts = "linneus{3,8,20}.inf.ethz.ch"

[linneus-upper]
hosts = "linneus{22-42}.inf.ethz.ch"

[linneus-mosix]
hosts = "linneus{48,49,51,56,57}.inf.ethz.ch"

[linneus-other]
hosts = "linneus{43,45-47,52,54,55,58-72}.inf.ethz.ch"

[linneus-new]
hosts = "linneus{43,71-78}.inf.ethz.ch"

[linneus-all]
#hosts=linneus-foo.inf.ethz.ch
groups = linneus-lower,linneus-upper,linneus-mosix,linneus-other,linneus-new

[bach]
hosts = "bach{05,06,12,18}.ethz.ch"

"""

__version__ = '0.1.2'

import locale
import logging
import os
import re
import subprocess
import sys

import configobj

import cliutils
from sans.ethz import Error
from sans.ethz import cli
from sans.ethz import util
from sans.ethz.cli import output

# Set the locale for all categories to the user’s default setting
locale.setlocale(locale.LC_ALL, '')

logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')

# Patch subprocess to have DEVNULL like in python >= 3.3
if not hasattr(subprocess, 'DEVNULL'):
    subprocess.DEVNULL = open(os.devnull, "w")


class HostCtl(cli.Command):
    """Usage:
        ${name} [ARGS...]
        ${name} COMMAND [ARGS...]
        ${name} help COMMAND

    Perform actions on one or more hosts.

    ${option_list}
    ${command_list}
    ${help_list}
    """

    name = 'hostctl'
    option_interactive = True

    _config = None

    @property
    def config(self):
        if not self._config:
            config_file = os.path.expanduser(self.options.config)
            if os.path.isfile(config_file):
                self._config = configobj.ConfigObj(config_file)
            else:
                self._config = {}
        return self._config

    def get_optparser(self):
        parser = super(HostCtl, self).get_optparser()
        parser.add_option('-c', '--config', type='string', dest='config', default='~/.hostctl.conf',
                          help='Path to config file, default: ~/.hostctl.conf')
        parser.add_option('-u', '--user', type='string', dest='user', default=None,
                          help='the username for authenticating')
        return parser

    # ---------- Commands ----------

    @cli.alias('ls')
    def do_list(self, subcmd, opts, *args):
        """list the machines I would operate on"""
        hosts = self._create_host_list(args)
        for h in hosts:
            print(h)

    def do_groups(self, subcmd, opts, *args):
        """list the configured groups"""
        if self.config:
            labels = ['Group', 'Hosts', 'Groups']
            entries = []
            for name, conf in self.config.items():
                if name == 'global':
                    continue
                entry = [name]
                hosts = conf.get('hosts', '')
                if isinstance(hosts, (list, tuple)):
                    hosts = ', '.join(hosts)
                entry.append(hosts)
                groups = conf.get('groups', '')
                if isinstance(groups, (list, tuple)):
                    groups = ', '.join(groups)
                entry.append(groups)
                entries.append(entry)
            output.print_table(entries, labels=labels)

    @cli.option('-t', '--timeout', dest='timeout', default='1', help='what to set ping -W to')
    def do_status(self, subcmd, opts, *args):
        """shows the status of a machine

        A machine is considered up if it answers to pings and a port scan
        reveils that a ssh server running."""
        hosts = self._create_host_list(args)

        def w(what):
            sys.stdout.write(what)
            sys.stdout.flush()

        ssh_re = re.compile('^22/tcp[ ]+(?P<state>closed|open)[ ]+ssh$', re.MULTILINE)
        ping_cmd = 'ping -c 1'.split(' ')
        ping_cmd.extend(('-W', opts.timeout))
        self.log.debug('ping_cmd: {0}'.format(ping_cmd))
        for host in hosts:
            self.log.debug('Pinging host: %s', host)
            w('{}: '.format(host))
            try:
                subprocess.check_call(ping_cmd + [host, ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                w('up')
                self.log.debug('Running nmap on host: %s', host)
                cmd = ['nmap', '-P0', '-p 22', host]
                p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                stdout, stderr = p.communicate()
                self.log.debug('nmap\'s stdout: %s', stdout.decode())
                match = ssh_re.search(stdout.decode())
                if match and match.group('state') == 'open':
                    w('; ssh up\n')
                else:
                    w('; ssh down\n')
            except subprocess.CalledProcessError:
                w('down\n')

    def do_shutdown(self, subcmd, opts, *args):
        """shutdown a machine"""
        hosts = self._create_host_list(args)
        for host in hosts:
            cmd = '{} "/sbin/shutdown -h now"'.format(host)
            if self.options.user:
                cmd = '-l {} {}'.format(self.options.user, cmd)
            cliutils.sh.ssh(cmd)

    def do_reboot(self, subcmd, opts, *args):
        """reboot a machine"""
        hosts = self._create_host_list(args)
        for host in hosts:
            cmd = '{} "/sbin/reboot"'.format(host)
            if self.options.user:
                cmd = '-l {} {}'.format(self.options.user, cmd)
            self.log.debug('running: ssh %s', cmd)
            cliutils.sh.ssh(cmd)

    def do_keyscan(self, subcmd, opts, *args):
        """get a machines ssh fingerprint
        and output it in a format suitable for a known_hosts file"""
        hosts = self._create_host_list(args)
        hostnames = ['{},{},{}'.format(h, h.partition('.')[0], util.get_ip_from_hostname(h)) for h in hosts]
        p = cliutils.sh.echo(['-e', '\n'.join(hostnames)]) | cliutils.Process('ssh-keyscan -T 2 -t rsa -f -')
        print(p)

    @cli.option('-n', '--number', dest='number', default=1, help='how many terminals to open for each hosts')
    def do_xterm(self, subcmd, opts, *args):
        """open a xterm with a ssh connection to the given machine

        ${cmd_option_list}
        """
        xterm = 'xterm'
        if 'global' in self.config:
            xterm = self.config['global'].get('xterm', xterm)
        hosts = self._create_host_list(args)
        for host in hosts:
            cmd = '-e ssh '
            if self.options.user:
                cmd = '{} -l {}'.format(cmd, self.options.user)
            for n in range(0, int(opts.number)):
                cliutils.sh[xterm]('{} {}'.format(cmd, host))

    def do_ping(self, subcmd, opts, *args):
        """open a xterm with a ping to the given machine

        ${cmd_option_list}
        """
        xterm = 'xterm'
        if 'global' in self.config:
            xterm = self.config['global'].get('xterm', xterm)
        hosts = self._create_host_list(args)
        for host in hosts:
            cliutils.sh[xterm]('-e ping {}'.format(host))

    #        cmd = [xterm, '-e', 'ping']
    #        processes = []
    #        import subprocess
    #        for host in hosts:
    # print 'cmd:', cmd
    #            p = subprocess.Popen(cmd + [host])
    #            processes.append(p)

    def do_cssh(self, subcmd, opts, *args):
        """run cssh with the given hostnames"""
        hosts = self._create_host_list(args)
        cmd = ['cssh']
        if self.options.user:
            cmd.append('--username')
            cmd.append(self.options.user)
        cmd.extend(hosts)
        self.log.debug('Running %s with: %s', subcmd, cmd)

        import subprocess
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        for line in iter(proc.stdout.readline, ''):
            print(line)

    @cli.option('-t', '--timeout', dest='timeout', help='what to set ssh option ConnectTimeout to')
    @cli.preserve_remaining_arguments
    def do_dsh(self, subcmd, opts, args, rargs):
        """run the given command on the given hostnames through dsh

        Everything after -- is passed unchanged as arguments to dsh.

        ${cmd_option_list}
        """
        assert rargs, 'Missing required arguments for dsh command'
        hosts = self._create_host_list(args)
        cmd = ['dsh']
        if opts.timeout:
            cmd.extend(('-o', '-o ConnectTimeout={0}'.format(opts.timeout)))
        # cmd.append('-v')
        cmd.append('-m')
        if self.options.user:
            cmd.append(','.join(['{}@{}'.format(self.options.user, h) for h in hosts]))
        else:
            cmd.append(','.join(hosts))
        cmd.extend(rargs)
        self.log.debug('Running %s with: %s', subcmd, cmd)

        import subprocess
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        for line in iter(proc.stdout.readline, ''):
            print(line)

    @cli.preserve_remaining_arguments
    def do_dig(self, subcmd, opts, args, rargs):
        """run dig for each of the given hostnames

        Everything after -- is passed unchanged as arguments to dig.
        """
        hosts = self._create_host_list(args)
        cmd = ['dig']
        # cmd.append('-v')
        if rargs:
            cmd.extend(rargs)
        import subprocess
        for host in hosts:
            host_cmd = cmd[:]
            host_cmd.append(host)
            self.log.debug('Running %s with: %s', subcmd, host_cmd)
            subprocess.Popen(host_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            output = subprocess.check_output(host_cmd)
            if output:
                for line in output.splitlines():
                    print(host + line.decode())
            else:
                print(host)

    @cli.preserve_remaining_arguments
    def do_screen(self, subcmd, opts, args, rargs):
        """run the given command on the given hosts in a screen session and detach from it.

        Everything after -- is passed unchanged as arguments to screen.
        Screen is run as: screen -U -d -m
        """
        assert rargs, 'Missing required arguments for screen command'
        hosts = self._create_host_list(args)
        cmd = ['ssh']
        # cmd.append('-v')
        cmd.append('-n')
        if self.options.user:
            cmd.append('-l')
            cmd.append(self.options.user)
        import subprocess
        for host in hosts:
            host_cmd = cmd[:]
            host_cmd.append(host)
            host_cmd.extend('screen -U -d -m'.split())
            host_cmd.extend(rargs)
            self.log.debug('Running %s with: %s', subcmd, host_cmd)
            proc = subprocess.Popen(host_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            for line in iter(proc.stdout.readline, ''):
                print(line)

    @cli.preserve_remaining_arguments
    def do_ipmi(self, subcmd, opts, args, rargs):
        """run ipmitool for each of the given hostnames

        Everything after -- is passed unchanged as arguments to ipmitool.
        ipmitool is run as: ipmitool -H $hostname
        """
        hosts = self._create_host_list(args)
        cmd = ['ipmitool']
        import subprocess
        for host in hosts:
            host_cmd = cmd[:]
            host_cmd.append('-H')
            host_cmd.append(host)
            host_cmd.extend(rargs)
            self.log.debug('Running %s with: %s', subcmd, host_cmd)
            proc = subprocess.Popen(host_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            for line in iter(proc.stdout.readline, ''):
                print('{}: {}'.format(host, line.rstrip('\n')))

    def do_powercycle(self, subcmd, opts, *args):
        """power cycle the given hostnames using ipmi.
        """
        password_file = os.path.expanduser('~/ethz/cbrg/ipmi-password')
        return self.do_ipmi('ipmi', opts, args, '-U ADMIN -f {0} power cycle'.format(password_file).split(' '))

    @cli.preserve_remaining_arguments
    def do_cdist(self, subcmd, opts, args, rargs):
        """run 'cdist config' on the given hosts

        Everything after -- is passed unchanged as arguments to cdist.
        """
        assert rargs, 'Missing required arguments for cdist command'
        hosts = self._create_host_list(args)
        cmd = ['cdist']
        cmd.extend(rargs)
        cmd.extend(hosts)
        self.log.debug('Running %s with: %s', subcmd, cmd)

        import subprocess
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        for line in iter(proc.stdout.readline, ''):
            print(line)

    @cli.alias('exit')
    @cli.alias('q')
    def _do_quit(self, argv):
        """hidden quit command to leave interactive shell"""
        sys.exit(0)

    # ---------- Help ----------

    def help_config(self):
        """configuration examples"""
        doc = """sar@icarus ~ $ cat ~/.hostctl.conf 
[global]
xterm = urxvt

[csb]
hosts = "csb{2,6,8-13}.inf.ethz.ch"

[cbrg]
hosts = "csb{2,7}.inf.ethz.ch"
groups = linneus-all

[linneus-lower]
hosts = "linneus{3,8,20}.inf.ethz.ch"

[linneus-upper]
hosts = "linneus{22-42}.inf.ethz.ch"

[linneus-mosix]
hosts = "linneus{48,49,51,56,57}.inf.ethz.ch"

[linneus-other]
hosts = "linneus{43,45-47,52,54,55,58-72}.inf.ethz.ch"

[linneus-new]
hosts = "linneus{43,71-78}.inf.ethz.ch"

[linneus-all]
hosts=linneus-test.inf.ethz.ch
groups = linneus-lower,linneus-upper,linneus-mosix,linneus-other,linneus-new

[bach]
hosts = "bach{02-12,18,!05-8}.ethz.ch"
      """
        doc = self._help_preprocess(doc, None)
        doc = doc.rstrip() + '\n'  # trim down trailing space
        self.stdout.write(self._str(doc))
        self.stdout.flush()

    def help_examples(self):
        """usage examples"""
        doc = """Examples:
    ${name} COMMAND ARGS

    ${name} list host{1,4-9,12-16,20}.domain.tld
    ${name} -u root shutdown @linneus-lower
    ${name} -u root reboot @some-group
    ${name} dsh host{!2,1-7,!5}.domain.tld -- /usr/bin/uptime
    ${name} cssh host{!2,1-7,!5}.domain.tld
      """
        # ${name} status host{1,2,5-7}.domain.tld
        # doc = self._help_reindent(doc)
        doc = self._help_preprocess(doc, None)
        doc = doc.rstrip() + '\n'  # trim down trailing space
        self.stdout.write(self._str(doc))
        self.stdout.flush()

    # ---------- Utilities ----------

    def _get_hosts_in_group(self, groupname):
        host_list = []
        if groupname in self.config:
            group_config = self.config[groupname]
        else:
            self.log.warn('Can not find group named \'%s\'', groupname)
            group_config = {}
        if 'hosts' in group_config:
            hosts = group_config['hosts']
            if not isinstance(hosts, list):
                hosts = [hosts]
            for host in hosts:
                host_list.extend(util.expand_range_pattern(host))
        if 'groups' in group_config:
            groups = group_config['groups']
            if not isinstance(groups, list):
                groups = [groups]
            for group in groups:
                host_list.extend(self._get_hosts_in_group(group))
        return host_list

    def _create_host_list(self, args):
        hosts = []
        if args:
            for arg in args:
                if arg.startswith('@'):
                    hosts.extend(self._get_hosts_in_group(arg[1:]))
                else:
                    hosts.extend(util.expand_range_pattern(arg))
        for host in hosts[:]:
            if host.startswith('!'):
                hosts.remove(host)
                if host[1:] in hosts:
                    hosts.remove(host[1:])
        return util.uniq(hosts)


def main():
    try:
        sys.exit(HostCtl().main())
    except Error as e:
        print(str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        # FIXME: if in interactive shell abort current operation instead of exiting
        sys.exit(0)


if __name__ == '__main__':
    main()
