#!/usr/bin/env python3
#
# Manage NAC entries by utilizing the web services at netcenter.ethz.ch.
#
# (c) 2009, Steven Armstrong, ETH Zurich
#

__version__ = '0.1.5'

import logging
import sys

from sans.ethz import cli, network, Error
from sans.ethz.api import netcenter
from sans.ethz.cli import input, output, util
from sans.ethz.util import get_username_password, httpclient

logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')


class NacCtl(cli.Command):
    """Usage:
        ${name} COMMAND [ARGS...]
        ${name} help COMMAND

    Manage NAC entries by utilizing the web services at netcenter.ethz.ch.

    ${option_list}
    ${command_list}
    ${help_list}
    """
    name = 'nacctl'
    option_interactive = True
    option_debug = True

    default_http_headers = {'Content-Type': 'text/xml'}

    _http = None

    @property
    def http(self):
        if not self._http:
            self._http = httpclient.get_http_client(username=self.options.user, host=netcenter.hostname)
        return self._http

    _netcenter = None

    @property
    def netcenter(self):
        if not self._netcenter:
            username, password = get_username_password(username=self.options.user, host=netcenter.hostname)
            self._netcenter = netcenter.Client(username=username, password=password)
        return self._netcenter

    def get_optparser(self):
        parser = super(NacCtl, self).get_optparser()
        parser.add_option('-u', '--user', type='string', dest='user', default=None,
                          help='the username for authenticating against webservices')
        return parser

    # ---------- Commands ----------

    _search_attributes = (
        'ip', 'hostname', 'id', 'mac', 'currentProfile', 'netSup', 'profiles', 'hostId', 'description')

    @cli.option('-x', '--xml', action='store_true', help='output raw xml')
    def do_search(self, subcmd, opts, *args):
        """search for NAC entries

        Usage:
            ${name} ${cmd_name} attr:value [attr:value attr:value]

        Where attributes are any combination of:
            ip               resolve mac by ip, then search by mac address
            hostname         resolve mac by hostname, then search by mac address
            id               search by nac id
            mac              search by mac address
            currentProfile   search by current profile
            netSup           search by net support group
            profiles         search by any profile
            hostId           search by hostid
            description      search by description

        Attribute names can be abbreviated as long as a unique prefix is used.

        ${cmd_option_list}
        """
        """
<query>
   <mac>%95</mac>
   <currentProfile>serv%</currentProfile>
   <netSup>id-%</netSup>
   <profiles>service-id%</profiles>
   <description>Xerox%</description>
   <macRaw>%1b%</macRaw>
   <hostId>foo.ethz.ch</hostId>
</query>

<list>
   <macMgmt id="148610">
       <mac>00-19-99-f6-2e-24</mac>
       <hostId>slaghitg183</hostId>
       <currentProfile>id-dock</currentProfile>
       <netSup>id</netSup>
       <lastSeen>2013-09-12 08:50</lastSeen>
       <changed>2013-01-10 07:23</changed>
       <changedBy>moussov</changedBy>
       <macRaw>001999f62e24</macRaw>
       <profiles>
            <profile id="21433">
                <profileName>id-dock</profileName>
            </profile>
            <profile id="11395">
                <profileName>service-id-stud</profileName>
            </profile>
       </profiles>
   </macMgmt>
</list>
        """
        assert args, 'Missing required arguments'
        args = list(args)
        properties = util.parse_properties(args, self._search_attributes, label='attribute')

        response, content = self._nac_search(properties)

        if opts.xml:
            print(content)
        else:
            tree = netcenter.get_tree_or_raise_error(response, content)
            records = list(tree)
            if len(records):
                labels = ['Id', 'MAC', 'netSup', 'profile', 'HostID']
                fields = ('@id', 'mac', 'netSup', 'currentProfile', 'hostId')
                output.print_xml_record_set(records, fields, labels=labels)

    _delete_attributes = ('id', 'mac', 'ip', 'hostname')

    def do_delete(self, subcmd, opts, *args):
        """delete a NAC entry

        Usage:
            ${name} ${cmd_name} attribute value

        Where attribute is one of:
            id               NAC id
            mac              mac address
            ip               resolve mac by ip, then delete based on mac address
            hostname         resolve mac by hostname, then delete based mac address

        Attribute names can be abbreviated as long as a unique prefix is used.
        """
        assert args, 'Missing required arguments: attribute, value'
        assert len(args) == 2, 'Wrong number of arguments. Expected 2, got %d' % len(args)
        attribute = util.get_canonical_name(args[0], self._delete_attributes, label='attribute')
        value = args[1]

        if attribute == 'id':
            _id = value
        else:
            _id = self._get_nac_id(attribute, value)

        uri = netcenter.uri('/netcenter/rest/nac/macMgmt/{0}'.format(_id))
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='DELETE', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)
        netcenter.get_tree_or_raise_error(response, content)
        # If we get here, deletion was successfull
        logging.info('Successfully deleted NAC entry %s', _id)

    _create_properties = ('ip', 'hostname', 'mac', 'description', 'hostId', 'currentProfile', 'netSup', 'profiles')

    def do_create(self, subcmd, opts, *args):
        """create a new NAC entry

        Usage:
            ${name} ${cmd_name} attr:value [attr:value attr:value]

        Where attributes are any combination of:
            ip               resolve mac by ip, then search by mac address
            hostname         resolve mac by hostname, then search by mac address
            mac*             mac address
            description      description
            hostId           optional hostid
            currentProfile   the current profile
            netSup           the net support group
            profiles         comma seperated list of profiles

        A * denotes a required field.

        Attribute names can be abbreviated as long as a unique prefix is used.
        You will be prompted to select values for missing required attributes.
        """
        """
        Usage:
            ${name} ${cmd_name} mac:value [isg:value currentProfile:value profiles:value,value,value]

# erfassen
POST /netcenter/rest/nac/macMgmt
<insert>
    <macMgmt>
        <mac>00-00-aa-ae-1b-95</mac>*
        <description>Xerox Phaser 4510DT</description>
        <hostId>foo-id</hostId>
        <currentProfile>service-id-vpp</currentProfile>
        <netSup>id-bd-vpp</netSup>*
        <profiles>
            <profile>
                <profileName>service-id-vpp</profileName>
            </profile>
        </profiles>
    </macMgmt>
</insert>
"""
        assert args, 'Missing required arguments'
        args = list(args)

        properties = util.parse_properties(args, self._create_properties)

        if 'mac' not in properties:
            for key in ('hostname', 'ip'):
                if key in properties:
                    value = properties.pop(key)
                    logging.info('resolving mac address for %s %s', key, value)
                    properties['mac'] = network.get_mac(value, netcenter_client=self.netcenter)
                    if 'hostId' not in properties:
                        properties['hostId'] = value
                    break

        assert 'mac' in properties, 'Missing required property \'mac\''

        # ensure keys exist
        properties.setdefault('netSup', '')
        properties.setdefault('currentProfile', '')
        # profiles must be a list
        if 'profiles' in properties:
            properties['profiles'] = [profile for profile in properties['profiles'].split(',') if profile]
        properties.setdefault('profiles', [])

        if not properties['netSup']:
            # prompt for isg
            isgs = self._nac_isgs()
            labels = ['Id', 'Name']
            isg = input.choose_one(isgs, fields=('id', 'name'), labels=labels,
                                   intro='Please select a net support group:')
            properties['netSup'] = isg['name']
            logging.info('Setting \'netSup\' to: %s', isg['name'])

        def select_profiles():
            # prompt for profiles
            profiles = list(self.netcenter.get_profile_info(properties['netSup']))
            _labels = ['Name', 'VPZ', 'netSup']
            profiles = input.choose_many(profiles, fields=('profil_name', 'vpz', 'profil_netsup'), labels=_labels,
                                         intro='Please select profiles for this NAC entry:')
            properties['profiles'] = [pr['profil_name'] for pr in profiles]
            logging.info('Setting \'profiles\' to: {0}'.format(properties['profiles']))

        def select_current_profile():
            # prompt for current profile
            profiles = [{'name': x} for x in properties['profiles']]
            _labels = ['Name']
            profile = input.choose_one(profiles, fields=('name',), labels=_labels,
                                       intro='Please select the currently active profile:')
            properties['currentProfile'] = profile['name']
            logging.info('Setting \'currentProfile\' to: %s', profile['name'])

        if not properties['profiles'] and not properties['currentProfile']:
            select_profiles()
            select_current_profile()

        if properties['currentProfile'] and not properties['currentProfile'] in properties['profiles']:
            logging.info('Appending \'currentProfile\' to profiles list.')
            properties['profiles'].append(properties['currentProfile'])

        if properties['profiles'] and not properties['currentProfile']:
            select_current_profile()

        if not properties['profiles']:
            select_profiles()

        logging.debug('properties: %s', properties)

        uri = netcenter.uri('/netcenter/rest/nac/macMgmt')
        headers = self.default_http_headers.copy()
        body = ['<insert><macMgmt>']
        for k, v in properties.items():
            if k in self._create_properties:
                if k == 'profiles':
                    body.append('<{}>'.format(k))
                    for p in v:
                        body.append('<profile><profileName>{}</profileName></profile>'.format(p))
                    body.append('</{}>'.format(k))
                else:
                    body.append('<{}>{}</{}>'.format(k, v, k))
        body.append('</macMgmt></insert>')
        body = ''.join(body)
        logging.debug('uri: %s', uri)
        logging.debug('body: %s', body)

        # TODO: error handling
        response, content = self.http.request(uri, method='POST', headers=headers, body=body,
                                              auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)
        tree = netcenter.get_tree_or_raise_error(response, content)
        # If we get here, deletion was successfull
        element = tree.find('id')
        nac_id = element.text
        logging.info('Successfully create NAC entry %s', nac_id)

    #    def do_profiles(self, subcmd, opts, *args):
    #        """list NAC profiles associated with the current user
    #
    #        ${cmd_option_list}
    #        """
    #        profiles = self._nac_profiles()
    #
    #        #labels = ['Id', 'Name', 'VPN', 'verify']
    #        #entries = [[x['id'], x['name'], x['remark'], x['vpnAccess'], x['hostVerify']] for x in profiles]
    #        labels = ['Id', 'Name', 'Email']
    #        entries = [[x['id'], x['name'], x['email']] for x in profiles]
    #        output.print_table(entries, labels=labels)

    @cli.option('-a', '--all', action='store_true', help='list all isgs')
    def do_isgs(self, subcmd, opts, *args):
        """list isgs

        By default the list is limited to isgs associated with the current user.

        ${cmd_option_list}
        """
        """
    <isg>
        <id>4136303</id>
        <name>inf-icos</name>
        <email>support@inf.ethz.ch</email>
        <vpzs>
            <vpz><name>inf-icos</name></vpz>

        </vpzs>
    </isg>

        """
        isgs = self._nac_isgs(opts.all)

        labels = ['Id', 'Name', 'email']
        entries = [[x['id'], x['name'], x['email']] for x in isgs]
        output.print_table(entries, labels=labels)

    @cli.alias('q')
    @cli.alias('exit')
    def _do_quit(self, argv):
        """hidden quit command to leave interactive shell"""
        sys.exit(0)

    # ---------- Utilities ----------

    @staticmethod
    def print_record_set(records, fields, labels=None):
        table = []
        for record in records:
            row = []
            for field in fields:
                row.append(str(record.find(field).text))
            table.append(row)
        output.print_table(table, labels=labels)

    def _load_nac_base_data(self):
        uri = netcenter.uri('/netcenter/nac/loadNacBaseData')
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = netcenter.get_tree_or_raise_error(response, content)
        return tree

    def _nac_profiles(self):
        """get a list of profiles from the webservice"""
        tree = self._load_nac_base_data()
        entries = []
        for record in tree.findall('profiles/profile'):
            entries.append({
                'id': record.get('id'),
                'vpzId': record.get('vpzId'),
                'profileName': record.find('profileName').text,
                'netSupId': record.find('netSupId').text,
                'vpzName': record.find('vpzName').text,
            })
        return entries

    def _nac_netsups(self):
        tree = self._load_nac_base_data()
        entries = []
        for record in tree.findall('netSups/netSup'):
            entries.append({
                'id': record.find('id').text,
                'name': record.find('name').text,
                'email': record.find('email').text,
            })
        return entries

    def _nac_isgs(self, list_all=False):
        """get a list of isgs from the webservice"""
        """
<isg>
    <id>4136303</id>
    <name>inf-icos</name>
    <email>support@inf.ethz.ch</email>
    <vpzs>
        <vpz><name>inf-icos</name></vpz>
    </vpzs>
</isg>
        """
        if list_all:
            uri = netcenter.uri('/netcenter/rest/user/listAllIsgs')
        else:
            uri = netcenter.uri('/netcenter/rest/user/listIsgByUser')
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = netcenter.get_tree_or_raise_error(response, content)

        entries = []
        for record in tree.findall('isg'):
            entries.append({
                'id': record.find('id').text,
                'name': record.find('name').text,
                'email': record.find('email').text,
            })
        return entries

    def _nac_search(self, properties):
        """query the webservice for a NAC entry"""
        if 'mac' not in properties:
            for key in ('hostname', 'ip'):
                if key in properties:
                    value = properties.pop(key)
                    logging.info('resolving mac address for %s %s', key, value)
                    properties['mac'] = network.get_mac(value, netcenter_client=self.netcenter)
                    break

        logging.debug('properties: {0}'.format(properties))

        uri = netcenter.uri('/netcenter/rest/nac/macMgmt/search')
        headers = self.default_http_headers.copy()
        body = ['<query>']
        for k, v in properties.items():
            if k in self._search_attributes:
                body.append('<{}>{}</{}>'.format(k, v, k))
        body.append('</query>')
        body = ''.join(body)
        logging.debug('uri: %s', uri)
        logging.debug('body: %s', body)

        response, content = self.http.request(uri, method='POST', headers=headers, body=body,
                                              auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)
        return response, content

    def _get_nac_id(self, attribute, value):
        """get a NAC entries ID"""
        response, content = self._nac_search({attribute: value})
        tree = netcenter.get_tree_or_raise_error(response, content)
        value = tree.find('macMgmt').get('id')
        return value

    # ---------- Help ----------
    # TODO: usage examples


def main():
    try:
        sys.exit(NacCtl().main())
    except netcenter.NetcenterAuthenticationError as e:
        print('Error: {}'.format(e))
        sys.exit(10)
    except Error as e:
        print('Error: {}'.format(e))
    except KeyboardInterrupt:
        # FIXME: if in interactive shell abort current operation instead of exiting
        sys.exit(0)


if __name__ == '__main__':
    main()
