#!/usr/bin/env python3
#
# Query the netcenter webservice for information about networks and hosts.
#
# (c) 2009, Steven Armstrong, ETH Zurich
#

__version__ = '0.1.5'

import logging
import re
import sys

from sans.ethz import cli, util
from sans.ethz.api import netcenter
from sans.ethz.cli import output

logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')


class NetInfo(cli.Command):
    """Usage:
        ${name} [NETINFO-OPTIONS...] [ARGS...]
        ${name} [NETINFO-OPTIONS...] COMMAND [COMMAND-SPECIFIC OPTIONS...] [ARGS...]
        ${name} [NETINFO-OPTIONS...] COMMAND [ARGS...] [COMMAND-SPECIFIC OPTIONS...]
        ${name} help COMMAND

    Query the netcenter webservice for information about networks and hosts.

    ${option_list}
    ${command_list}
    ${help_list}
    """
    name = 'netinfo'
    option_interactive = True

    _netcenter = None

    @property
    def netcenter(self):
        if not self._netcenter:
            username, password = util.get_username_password(username=self.options.user, host=netcenter.hostname)
            self._netcenter = netcenter.Client(username=username, password=password)
        return self._netcenter

    def get_optparser(self):
        parser = super(NetInfo, self).get_optparser()
        parser.add_option('-u', '--user', type='string', dest='user', default=None,
                          help='the username for authenticating against the webservice')
        parser.add_option('-e', '--extra-fields', type='string', dest='extra', default=None,
                          help='comma seperated list of extra fields to output in results, e.g. -e \'last_seen,switch_ip\'')
        return parser

    # ---------- Commands ----------

    def do_ip(self, subcmd, opts, *args):
        """search by ip address"""
        self.log.info('Searching by ip: %s', args)
        results = self.netcenter.search_by_ip_name_mac_cable(args)
        self.process_results(results)

    def do_host(self, subcmd, opts, *args):
        """search by hostname"""
        self.log.info('Searching by hostname: %s', args)
        results = self.netcenter.search_by_ip_name_mac_cable(args)
        self.process_results(results)

    def do_mac(self, subcmd, opts, *args):
        """search by MAC address"""
        # netcenter suddenly only accepts dash delimited mac addresses ...
        args = [util.format_mac(arg, '-') for arg in args]
        self.log.info('Searching by mac address: %s', args)
        results = self.netcenter.search_by_ip_name_mac_cable(args)
        self.process_results(results)

    def do_cable(self, subcmd, opts, *args):
        """search by cable, UTP or PAT"""
        self.log.info('Searching by cable: %s', args)
        results = self.netcenter.search_by_ip_name_mac_cable(args)
        self.process_results(results)

    @cli.option('-r', '--room', metavar='ROOM', help='limit search to the given rooms. Comma seperated list.')
    def do_vlan(self, subcmd, opts, *args):
        """search by vlan

        Usage:
            ${name} ${cmd_name} [OPTION]

        ${cmd_option_list}
        """
        self.log.info('Searching by vlan: %s', args)
        if opts.room:
            opts.room = opts.room.split(',')
            self.log.info('  filtering by room: %s', opts.room)
        results = self.netcenter.search_by_vlan(args, rooms=opts.room)
        self.process_results(results)

    @cli.option('-v', '--vlan', metavar='VLAN', help='limit search to the given vlans. Comma seperated list.')
    def do_room(self, subcmd, opts, *args):
        """search by room

        Usage:
            ${name} ${cmd_name} [OPTION]

        ${cmd_option_list}
        """
        self.log.info('Searching by room: %s', args)
        if opts.vlan:
            opts.vlan = opts.vlan.split(',')
            self.log.info('  filtering by vlan: %s', opts.vlan)
        results = self.netcenter.search_by_room(args, vlans=opts.vlan)
        self.process_results(results)

    @cli.option('-6', '--ipv6', action='store_true', help='include IPv6 subnets')
    def do_subnet(self, subcmd, opts, *args):
        """search for subnet info by subnet, ip or vlan

        Usage:
            ${name} ${cmd_name} [OPTION]

        ${cmd_option_list}
        """
        self.log.info('Searching for subnet: %s', args)
        results = list(self.netcenter.get_subnet_info(args))
        if self.options.quiet:
            labels = None
        else:
            labels = ('VLAN', 'VLAN Name', 'VNZ', 'Profiles', 'IPv4 Subnets')
            if opts.ipv6:
                labels += ('IPv6 Subnets',)
        fields = ('vlan', 'vlan_name', 'vnz', 'profiles', 'ip_subnets')
        if opts.ipv6:
            fields += ('ipv6_subnets',)

        # strip out useless info
        def shorten(newline_delimited_list):
            out = []
            for item in newline_delimited_list.split('\n'):
                out.append(item.split(' ')[0])
            return '\n'.join(out)

        for entry in results:
            entry['profiles'] = shorten(entry['profiles'])
            entry['ip_subnets'] = shorten(entry['ip_subnets'])
            if opts.ipv6:
                entry['ipv6_subnets'] = shorten(entry['ipv6_subnets'])
        output.print_dict_record_set(results, fields, labels=labels)

    @cli.alias('q')
    @cli.alias('exit')
    def _do_quit(self, argv):
        """hidden quit command to leave interactive shell"""
        sys.exit(0)

    # ---------- Help ----------

    def help_examples(self):
        """usage examples"""
        doc = """Examples:
    ${name} COMMAND ARGS

    ${name} ip 129.132.12.88
    ${name} -e last_seen,switch_ip ip 129.132.12.88
    ${name} host foo.inf.ethz.ch
    ${name} mac 00:1B:FC:A2:C3:9F
    ${name} cable UTP129992
    ${name} vlan 781
    ${name} vlan -r HPT/D/16 745
    ${name} vlan 745 -r HPT/D/16
    ${name} room CAB/D/56

If COMMAND is ommited it is guessed by inspecting the arguments:
    ${name} 129.132.12.88 -> ${name} ip 129.132.12.88
    ${name} -e last_seen 129.132.12.88 -> ${name} -e last_seen ip 129.132.12.88
    ${name} foo.inf.ethz.ch -> ${name} host foo.inf.ethz.ch
      """
        doc = self._help_reindent(doc)
        doc = self._help_preprocess(doc, None)
        doc = doc.rstrip() + '\n'  # trim down trailing space
        self.stdout.write(self._str(doc))
        self.stdout.flush()

    # ---------- Overriden methods ----------

    _autodetect_patterns = {
        # 127.0.0.1
        'ip': re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'),
        # host.ethz.ch | ethz.ch
        'host': re.compile(r'^(?:[a-z0-9][a-z0-9-]{0,61}[a-z0-9]\.)+[a-z]{2,6}$', re.IGNORECASE),
        # 00:1B:FC:A2:C3:9F | 00-1B-FC-A2-C3-9F
        'mac': re.compile(r'^(?:[a-f0-9]{2}[\:-]){5}[a-f0-9]{2}$', re.IGNORECASE),
        # UTP129992 | PAT137051
        'cable': re.compile(r'^(?:UTP|PAT)\d{6}$', re.IGNORECASE),
        # 781 | Cx::781
        'vlan': re.compile(r'^(?:[^\:]+\:{2})?\d{3}$', re.IGNORECASE),
        # CAB/E/48
        'room': re.compile(r'^[a-z]{2,3}/[a-z]{1}/\d+(?:.\d+)*$', re.IGNORECASE),
    }

    def precmd(self, argv):
        """If no explicit command name was given try guessing it by inspecting the arguments"""
        if argv:
            cmdname = self._get_canonical_cmd_name(argv[0])
            if cmdname:
                args = argv[1:]
            else:
                args = argv
            # Expand range patterns here for all commands
            #  @see ethz.util.expand_range_pattern
            args = self._create_argument_list(args)
            if not cmdname:
                self.log.debug('No command given. Autodetecting based on: %s', args)
                # arg = argv[0]
                arg = args[0]
                # arg = util.expand_range_pattern(argv[0])[0]
                for name, pattern in self._autodetect_patterns.items():
                    if pattern.search(arg):
                        cmdname = name
                        # args.insert(0, name)
                        break
            if cmdname:
                self.log.debug('Calling command: %s; with arguments: %s', cmdname, args)
                argv = [cmdname]
                argv.extend(args)
        return argv

    # ---------- Utilities ----------

    @staticmethod
    def _create_argument_list(args):
        args_ = []
        if args:
            for arg in args:
                args_.extend(util.expand_range_pattern(arg))
        return util.uniq(args_)

    def process_results(self, results):
        for entry in results:
            try:
                self.log.debug('entry: %s', entry)
                _format = []
                if entry.get('end_location', ''):
                    _format.append('%(end_location)s')
                if entry.get('cable_id', ''):
                    _format.append('%(cable_id)s')
                elif entry.get('switch_name', ''):
                    _format.append('%(switch_name)s/%(switch_port)s')

                if netcenter.is_phone(entry):
                    _format.append('TEL')
                    _format.append('%(i_phone_no)s')
                elif netcenter.is_host(entry):
                    _format.append('NET')
                    _format.append('%(vlan)s')
                    if 'ip_name' in entry:
                        _format.append('%(ip_name)s')
                    if 'ip' in entry:
                        _format.append('%(ip)s')
                    if 'mac' in entry:
                        _format.append('%(mac)s')
                elif netcenter.is_notused(entry):
                    _format.append('NOTUSED')
                else:
                    _format.append('UNKNOWN')
                if self.options.extra:
                    for k in self.options.extra.split(','):
                        if k in entry:
                            _format.append('%%(%s)s' % k)
                print(' '.join(_format) % entry)
            except KeyError as e:
                print('Error: {} {}'.format(e, entry))


def main():
    try:
        sys.exit(NetInfo().main())
    except netcenter.NetcenterAuthenticationError as e:
        print('Error: {}'.format(e))
        sys.exit(10)
    except netcenter.NetcenterError as e:
        print('Error: {}'.format(e))
        sys.exit(1)
    except KeyboardInterrupt:
        # FIXME: if in interactive shell abort current operation instead of exiting
        sys.exit(0)


if __name__ == '__main__':
    main()
