import logging
import re
import socket

from sans.ethz import Error, util
from sans.ethz.api import netcenter


class NetworkError(Error):
    """Base exception for this module"""
    pass


_netcenter = None


def get_netcenter_client():
    global _netcenter
    if not _netcenter:
        username, password = util.get_username_password(username=None, host=netcenter.hostname)
        _netcenter = netcenter.Client(username=username, password=password)
    return _netcenter


mac_re = re.compile(r'^(?:[a-f0-9]{2}[:-]){5}[a-f0-9]{2}$', re.IGNORECASE)


def is_mac_address(mac):
    if mac_re.search(mac):
        return True
    else:
        return False


def get_host_info(ip_name_mac_cable, netcenter_client=None):
    kc = netcenter_client or get_netcenter_client()
    hosts = list(kc.get_host_info(ip_name_mac_cable))
    if not hosts:
        raise netcenter.NetcenterError('Failed to get host info for {}'.format(ip_name_mac_cable))
    return hosts[0]


def get_mac(hostname_or_ip, netcenter_client=None):
    try:
        host = get_host_info(hostname_or_ip, netcenter_client=netcenter_client)
        if host:
            return host['mac']
    except netcenter.NetcenterAuthenticationError:
        raise
    except netcenter.NetcenterError as e:
        logging.warning(e)
    raise NetworkError('Failed to get MAC address for \'{}\' from netcenter.'.format(hostname_or_ip))


def get_ip(hostname_or_mac, netcenter_client=None):
    try:
        host = get_host_info(hostname_or_mac, netcenter_client=netcenter_client)
        if host:
            return host['ip']
    except netcenter.NetcenterAuthenticationError:
        raise
    except netcenter.NetcenterError as e:
        logging.warning(e)

    if is_mac_address(hostname_or_mac):
        raise NetworkError('Failed to get IP for MAC \'{}\' from netcenter.'.format(hostname_or_mac))
    else:
        try:
            # Check the DNS database
            kc = netcenter_client or get_netcenter_client()
            ip = kc.get_host_ip(hostname_or_mac)
            if ip:
                return ip
        except netcenter.NetcenterAuthenticationError:
            raise
        except netcenter.NetcenterError as e:
            logging.warning(e)

        logging.warning('Failed to get IP for hostname \'%s\' from netcenter. Falling back to socket module.',
                        hostname_or_mac)
        return get_ip_from_hostname(hostname_or_mac)


def get_hostname(ip_or_mac, netcenter_client=None):
    try:
        host = get_host_info(ip_or_mac, netcenter_client=netcenter_client)
        if host:
            return host['ip_name']
    except netcenter.NetcenterAuthenticationError:
        raise
    except netcenter.NetcenterError as e:
        logging.warning(e)

    if is_mac_address(ip_or_mac):
        raise NetworkError('Failed to get hostname for MAC \'{}\' from netcenter.'.format(ip_or_mac))
    else:
        logging.warning('Failed to get hostname for IP \'%s\' from netcenter. Falling back to socket module.',
                        ip_or_mac)
        return get_hostname_from_ip(ip_or_mac)


def get_ip_from_hostname(hostname):
    try:
        return socket.gethostbyname(hostname)
    except socket.gaierror as e:
        raise NetworkError('Failed to get IP for hostname \'{}\': {}'.format(hostname, e))


def get_hostname_from_ip(ip):
    try:
        return socket.gethostbyaddr(ip)[0]
    except socket.gaierror as e:
        raise NetworkError('Failed to get hostname for IP \'{}\': {}'.format(ip, e))
    except IndexError as e:
        raise NetworkError('Failed to get hostname for IP \'{}\': {}'.format(ip, e))
