import logging
import os
import re
import string
import sys


def uniq(l):
    """make a list uniq while preserving order"""
    s = {}
    return [s.setdefault(e, e) for e in l if e not in s]


def _expand_range(s):
    """_expand_range(range_pattern) -> list of strings"""
    included = []
    excluded = []
    parts = s.split(',')
    for segment in parts:
        if segment.startswith('!'):
            segment = segment[1:]
            collection = excluded
        else:
            collection = included
        delim = None
        if '-' in segment:
            delim = '-'
        elif '..' in segment:
            # bash style
            delim = '..'
        if delim:
            start, end = segment.split(delim, 1)
            if start.startswith('0'):
                # pad with leading zeros
                _format = '%%0%dd' % len(start)
            else:
                _format = '%d'
            collection.extend([_format % i for i in range(int(start), int(end) + 1)])
        else:
            collection.append(segment)
    _range = [x for x in included if x not in excluded]
    return _range


def _merge_range(prefix, _range, suffix):
    """_merge_range(prefix, range_pattern, suffix) -> list of strings"""
    return ['%s%s%s' % (prefix, r, suffix) for r in _expand_range(_range)]


range_re = re.compile(r'([^{]*){([^{]*)}([^{]*)', re.IGNORECASE)


def expand_range_pattern(s):
    """expand_range_pattern(pattern) -> list of strings
    
    >>> expand_range_pattern('a{1-3}b')
    ['a1b', 'a2b', 'a3b']
    
    >>> expand_range_pattern('a{5..6}b')
    ['a5b', 'a6b']
    
    >>> expand_range_pattern('a{1,2,3-5}b')
    ['a1b', 'a2b', 'a3b', 'a4b', 'a5b']
    
    pad with leading zeros
    >>> expand_range_pattern('a{01-3}b')
    ['a01b', 'a02b', 'a03b']
    
    >>> expand_range_pattern('a{1-3}b{4,5}')
    ['a1b4', 'a1b5', 'a2b4', 'a2b5', 'a3b4', 'a3b5']

    >>> expand_range_pattern('a{1-5,!2-3}b{4,5}')
    ['a1b4', 'a1b5', 'a4b4', 'a4b5', 'a5b4', 'a5b5']

    >>> expand_range_pattern('a{1-3}b{4,5,!4}')
    ['a1b5', 'a2b5', 'a3b5']

    """
    matches = range_re.findall(s)
    if not matches:
        # assume just a single hostname
        return [s]
    else:
        names = []
        # popping from end of list is more efficient
        matches.reverse()
        while matches:
            parts = _merge_range(*matches.pop())
            if not names:
                names = parts
                continue
            tmp = []
            for name in names:
                for part in parts:
                    tmp.append('%s%s' % (name, part))
            names = tmp
        return names


# s = 'a{1-3}b{5..6}c{8-10,12}d'
# for n in expand_range_pattern(s):
#    print n


def get_ip_from_hostname(hostname):
    import socket
    import sys
    ipaddress = None
    try:
        try:
            ipaddress = socket.gethostbyname(hostname)
        except OSError as ex:
            print('get_ip_from_hostname: {}'.format(ex), file=sys.stderr)
    finally:
        return ipaddress


def get_username_password(username=None, host=None):
    logging.debug('get_username_password: username={0}, host={1}'.format(username, host))

    # check for username/password environment variables
    password = os.getenv('SANS_PASSWORD')
    # for username, a command line provided value has precedence
    if not username:
        username = os.getenv('SANS_USERNAME')

    if not username:
        # Take username from command line, prompt for password
        if host:
            try:
                import netrc
                n = netrc.netrc()
                authenticators = n.authenticators(host)
                if authenticators:
                    # Use credentials found in ~/.netrc file
                    username, password = authenticators[0], authenticators[2]
            except IOError:
                # no ~/.netrc file
                pass
        if not username:
            # Use username of the current user, prompt for password
            import getpass
            username = getpass.getuser()
    if not password:
        import getpass
        prompt = 'Password for %s' % username
        if host:
            prompt = '%s@%s' % (prompt, host)
        try:
            password = getpass.getpass(prompt='%s: ' % prompt, stream=sys.stderr)
        except TypeError:  # unix_getpass() got an unexpected keyword argument 'stream'
            # Fallback for old python version
            password = getpass.getpass(prompt='%s: ' % prompt)

    return username, password


not_hexdigits = re.compile(r'[^%s]' % string.hexdigits)


def format_mac(mac, delimiter=':'):
    if not mac:
        return mac

    # nuke any existing delimiters
    mac = not_hexdigits.sub('', mac)

    # /me likes uppercase mac addresses
    mac = mac.upper()

    # re-format with given delimiter
    return delimiter.join([mac[i:i + 2] for i in range(0, len(mac), 2)])
